package org.gregoire.portknock.client;

import java.util.*;
import java.io.*;
import java.net.*;

/**
 * A client that attempts connections to a given server following a sequence of
 * ports to open within a specific time period.
 * <br>
 * This is test implementation
 *
 * @author Paul Gregoire <paul%40gregoire.org>
 * @version 0.0.1
 */
public class Main {

    private String host;
    private int socketTimeout;
    private int delay;
    private int port;
    private int[] sequencePorts;

    /**
     * Initializes all the internal members.
     *
     * @param props from a client configuration
     */
    public void init(Properties props) {
        //get number of ports used for sequence
        int sequenceLength = Integer.parseInt(props.getProperty("good.ports.sequence"));
        //get ports used for sequence
        String tmp = props.getProperty("good.ports");
        StringTokenizer st = new StringTokenizer(tmp, ",");
        if (st.countTokens() < sequenceLength) {
            sequenceLength = st.countTokens();
        }
        //setup an array to hold the ports
        sequencePorts = new int[sequenceLength];
        for (int i = 0; i < sequenceLength; i++) {
            sequencePorts[i] = Integer.parseInt(st.nextToken());
        }

        //the time to wait in between connection attempts
        delay = Integer.parseInt(props.getProperty("connection.delay"));

        //get host name or ip address used for sequence
        host = props.getProperty("host.address");

        //the last port that will be opened for connection if sequence is ok
        port = Integer.parseInt(props.getProperty("good.ports.final"));

        //get the default socket timeout
        socketTimeout = Integer.parseInt(props.getProperty("socket.timeout"));

    }

    /**
     * Returns the socket to be used for the tunneling application. The socket
     * is only returned if the sequence is successful.
     *
     * @return socket available if sequence is successful.
     * @throws IOException
     */
    public Socket connect() throws IOException {
        //try all the sequence ports in order
        for (int i = 0; i < sequencePorts.length; i++) {
            Socket sock = new Socket(host, sequencePorts[i]);
            sock.setKeepAlive(false);
            sock.setReuseAddress(true);
            sock.setSoTimeout(socketTimeout);
            //close it down
            sock.close();
            try {
                Thread.currentThread().sleep(delay);
            } catch (InterruptedException ex) {
            }
            System.out.println("Done with: " + sequencePorts[i]);
        }
        System.out.println("Sequence completed");
        //try to open the final port..
        Socket sockFinal = new Socket(host, port);
        sockFinal.setKeepAlive(true);
        sockFinal.setReuseAddress(true);
        sockFinal.setSoTimeout(socketTimeout);
        //return the socket if we dont throw an exception
        return sockFinal;
    }

    public static void main(String args[]) {
        Socket socket = null;
        Properties props = new Properties();
        try {

            props.load(new FileInputStream("client.properties"));

            try {

                Main main = new Main();
                main.init(props);
                socket = main.connect();
                if (socket.isConnected()) {
                    System.out.println("Socket returned and connected...");
                    byte[] buff = new byte[1024];
                    InputStream is = socket.getInputStream();
                    while (is.read(buff) != -1) {
                        System.out.print(new String(buff));
                    }
                    System.out.println();
                    is.close();
                }

            } catch (IOException ex) {
                System.err.println("Error with connection.");
            } finally {
                try {
                    if (null != socket) {
                        socket.close();
                    }
                } catch (Exception ex) {
                }
            }

        } catch (IOException ex) {
            System.err.println("Error loading properties file, server will exit.");
        }
        //normal exit
        System.exit(0);

    }

}
