package org.gregoire.portknock.server;

import java.io.*;
import java.net.*;
import java.nio.*;
import java.util.*;

import org.apache.log4j.*;
import com.ibm.io.async.*;

/**
 * Generic AsyncIO server
 *
 * @author Paul Gregoire <paul%40gregoire.org>
 * @version 0.0.1
 */
public class AsyncServerImpl extends Server {

    private static int counter = 0;

    private AsyncServerSocketChannel ssc;
    private AsyncServerSocketChannel finalChannel;

    protected LinkedList nextChannelList;

    static {
        //log for this class only
        logger = Logger.getLogger(AsyncServerImpl.class.getName());
    }

    public AsyncServerImpl(String type) {
        super(type);
        //list to hold all the "connectors"
        list = new LinkedList();
        //list to hold all the "next" channels
        nextChannelList = new LinkedList();
    }

    public void init() throws IOException {
        // Create a new server socket and set the blocking mode
        ssc = AsyncServerSocketChannel.open();
    }

    // Accept connections for current server.
    public void listen() {
        try {
            ServerSocket serverSocket = ssc.socket();
            // Set socket options.
            serverSocket.setReuseAddress(true);
            // Bind the server socket to the local host and port or the incomming host and port
            InetSocketAddress isa = new InetSocketAddress(host, ports[0]);
            serverSocket.bind(isa);

            logger.debug("In listen, so timeout: " + serverSocket.getSoTimeout());

            // Listen for client connections.
            AsyncSocketChannel clientChannel = ssc.accept();
            //set the timeout
            //ssc.socket().setSoTimeout(timeout);
            if (clientChannel.isConnected()) {
                // The key indexes into the selector so you
                // can retrieve the socket that's ready for I/O
                SocketHandler handler = new SocketHandler();
                Socket socket = clientChannel.socket();
                handler.setSocket(socket);
                new Thread(handler, "Socket handler " + counter++).start();
                socket = null;
            }
        } catch (Exception ex) {
            logger.log(Level.ERROR, "Exception in listen: " + ex.getMessage(), ex);
        }

    }

    public class SocketHandler implements Runnable {

        Socket socket = null;

        public void setSocket(Socket socket) {
            this.socket = socket;
        }

        public void run() {
            try {
                handle();
                socket.close();
            } catch (IOException ex) {
            }
        }

        /**
         * Given a Socket, this method builds the objects necessary to communicate with
         * the broadcaster and begins receiving the stream
         *
         * @param socket The socket on which the client has connected
         * @throws java.io.IOException
         */
        private void handle() throws IOException {
            InetAddress address = socket.getInetAddress();
            Client client = new Client(address);
            if (!list.contains(client)) {
                logger.log(Level.DEBUG, "Client not in list");
                list.add(client);
            } else {
                Client tmp = null;
                int listSize = list.size();
                for (int c = 0; c < listSize; c++) {
                    tmp = (Client) list.get(c);
                    if (client.equals(tmp)) {
                        //set the reference to the current client
                        client = tmp;
                        //increment the connection count
                        client.incrementSequence();
                        //if we have reached the end of the sequence then open their tunnel
                        if (client.getSequence() == ports.length) {
                            //remove the record
                            list.remove(c);
                            //open tunnel and allow only current requester in
                            logger.log(Level.DEBUG, "Got a good sequence");
                            listenFinal(client);
                            //return control
                            return;
                        }
                        break;
                    }
                    tmp = null;
                }
            }

            logger.log(Level.DEBUG, socket.getInetAddress().toString());
            logger.log(Level.DEBUG, "List size: " + list.size());

            //listen on the next port in the sequence
            listenNext(client);

        }

        // Accept connections on a follow on sequence port
        public void listenNext(Client client) {
            int sequence = client.getSequence();
            logger.log(Level.DEBUG, "Current port: " + ports[sequence]);
            ServerSocket serverSocket = null;
            AsyncServerSocketChannel nextChannel = null;
            AsyncSocketChannel clientChannel = null;
            try {
                if (nextChannelList.size() >= sequence) {
                    //look for a matching channel in the channel list
                    Object tmp = nextChannelList.get(sequence - 1);
                    if (null != tmp) {
                        nextChannel = (AsyncServerSocketChannel) tmp;
                    }
                    //bind the socket channel to the inet address
                    serverSocket = nextChannel.socket();
                } else {
                    //create channel for "next" sockets
                    nextChannel = AsyncServerSocketChannel.open();
                    //add it to the list
                    nextChannelList.add( (sequence - 1), nextChannel);
                    //bind the socket channel to the inet address
                    serverSocket = nextChannel.socket();
                    // Set socket options.
                    serverSocket.setReuseAddress(true);
                    //set the timeout
                    serverSocket.setSoTimeout(portDelay);
                    // Bind the server socket to the local host and port or the incomming host and port
                    InetSocketAddress isa = new InetSocketAddress(host, ports[sequence]);
                    // Bind the server socket to the local host and port or the incomming host and port
                    serverSocket.bind(isa);
                }
                while (true) {
                    // Listen for client connections.
                    clientChannel = nextChannel.accept();
                    //check to see if this new "connector" is the one we are expecting
                    if (!client.getAddress().equals(clientChannel.socket().getInetAddress())) {
                        logger.log(Level.DEBUG, "Got a bad connect");
                    } else {
                        logger.log(Level.DEBUG, "Got a follow on connect");
                        handle();
                        break;
                    }
                }
            } catch (Exception ex) {
                logger.log(Level.ERROR, "Exception in listen next: " + ex.getMessage(), ex);
            } finally {
                try {
                    clientChannel.close();
                } catch (IOException ex1) {
                    logger.log(Level.ERROR, "Error closing server socket", ex1);
                }
            }

        }

        // Accept connections on a follow on sequence port
        public void listenFinal(Client client) {
            logger.log(Level.DEBUG, "Current port: " + port);
            ServerSocket serverSocket = null;
            AsyncSocketChannel clientChannel = null;
            try {
                //if final channel is null then open it
                if (null == finalChannel) {
                    //create channel for "final" sockets
                    finalChannel = AsyncServerSocketChannel.open();
                }
                //get the socket channel
                serverSocket = finalChannel.socket();
                // Set socket options.
                serverSocket.setReuseAddress(true);
                //set the timeout
                serverSocket.setSoTimeout(portDelay);
                // Bind the server socket to the local host and port or the incomming host and port
                InetSocketAddress isa = new InetSocketAddress(host, port);
                // Bind the server socket to the local host and port or the incomming host and port
                serverSocket.bind(isa);
                while (true) {
                    // Listen for client connections.
                    clientChannel = finalChannel.accept();
                    //check to see if this new "connector" is the one we are expecting
                    if (!client.getAddress().equals(clientChannel.socket().getInetAddress())) {
                        logger.log(Level.DEBUG, "Got a bad final connect");
                    } else {
                        logger.log(Level.DEBUG, "Got the final connect");
                        //now dump out a bunch of "ok"'s to the out stream
                        if (type.equals("sequence")) {
                            //do stuff with socket
                            ByteBuffer buffer = ByteBuffer.allocateDirect(1024);
                            buffer.put("ok ok ok ok ok ok\n".getBytes());
                            buffer.flip();
                            IAsyncFuture writeFuture = clientChannel.write(buffer);
                            try {
                                logger.debug("Bytes written: " + writeFuture.getByteCount());
                            } catch (AsyncException ex) {
                                logger.error(ex);
                            }

                            /*
                               reading
                                ByteByffer buffer = ByteBuffer.allocateDirect(1024);
                                IAsyncFuture readFuture = channel.read(buffer);
                                ... <do some useful work>...
                                try {
                                   long bytesRead = readFuture.getByteCount();
                                    ...
                               } catch (AsyncException ex) {
                                   ... the IO operation failed...
                               }
                             */

                        } else if (type.equals("tunnel")) {
                            String path = client.getRequestedApplication();
                            String env[] = {
                                "JAVA_HOME=E:/j2sdk1.4.2_05"};
                            Process p = Runtime.getRuntime().exec(path, env);

                            // hook processes input to browser's output (async)
                            final InputStream inFromReq = socket.getInputStream();
                            final OutputStream outToCgi = socket.getOutputStream();
                            final byte[] txBuffer = new byte[2048];

                            new Thread(new Runnable() {
                                public void run() {
                                    try {
                                        while (inFromReq.read(txBuffer) > -1) {
                                            outToCgi.write(txBuffer);
                                            outToCgi.flush();
                                        }
                                        outToCgi.close();
                                    } catch (IOException e) {
                                        logger.warn(e);
                                    }
                                }
                            }).start();
                        }
                        break;
                    }
                }
            } catch (Exception ex) {
                logger.log(Level.ERROR, "Exception in listen next: " + ex.getMessage(), ex);
            } finally {
                try {
                    clientChannel.close();
                } catch (IOException ex1) {
                    logger.log(Level.ERROR, "Error closing socket", ex1);
                }
            }

        }

    }

}
