package org.gregoire.portknock.server;

import java.io.*;
import java.util.*;
import org.apache.log4j.*;

/**
 * Server entry point / initializer
 *
 * @author Paul Gregoire
 */
public class Main {

    private static boolean terminate = false;

    //this should be the first instance of the logger
    private static Logger logger = null;

    public void init(Properties props) throws Exception {

        String file = props.getProperty("log.config");
        // if the log4j-init-file is not set, then no point in trying
        if (file != null) {
            PropertyConfigurator.configure(file);
        }

        //log for this class only
        logger = Logger.getLogger(Main.class.getName());
        logger.log(Level.INFO, "Server initializing");

        //get number of ports used for sequence
        int sequenceLength = Integer.parseInt(props.getProperty("good.ports.sequence"));
        //get ports used for sequence
        String tmp = props.getProperty("good.ports");
        StringTokenizer st = new StringTokenizer(tmp, ",");
        if (st.countTokens() < sequenceLength) {
            sequenceLength = st.countTokens();
        }
        //setup an array to hold the ports
        int[] ports = new int[sequenceLength];
        for (int i = 0; i < sequenceLength; i++) {
            ports[i] = Integer.parseInt(st.nextToken());
        }

        //last port to open if sequence is completed
        int finalPort = Integer.parseInt(props.getProperty("good.ports.final"));

        //total time that a sequence can take
        int timeLimit = Integer.parseInt(props.getProperty("good.ports.timelimit"));
        int portDelay = (timeLimit / sequenceLength);

        //get number of ports used for deception
        sequenceLength = Integer.parseInt(props.getProperty("bad.ports.open"));
        int[] badPorts = null;
        //any number less than one will disable deceptive ports
        if (sequenceLength > 0) {
            //get ports used for deception
            tmp = props.getProperty("bad.ports");
            st = new StringTokenizer(tmp, ",");
            if (st.countTokens() < sequenceLength) {
                sequenceLength = st.countTokens();
            }
            //setup an array to hold the ports
            badPorts = new int[sequenceLength];
            for (int i = 0; i < sequenceLength; i++) {
                badPorts[i] = Integer.parseInt(st.nextToken());
            }
        }

        //get host name or ip address used for sequence
        String host = props.getProperty("host.address");

        //get the default socket timeout
        int socketTimeout = Integer.parseInt(props.getProperty("socket.timeout"));

        //get the number of handler threads
        //int requestHanders = Integer.parseInt(props.getProperty("handler.threads"));

        //get server type implementation to use
        String serverType = props.getProperty("server.type");

        //setup the server to handle the sequences
        Server sequenceHandler = null;
        if (serverType.equals("nio")) {
            sequenceHandler = new NIOServerImpl("sequence");
        } else if (serverType.equals("aio")) {
            sequenceHandler = new AsyncServerImpl("sequence");
        } else {
            throw new Exception("Server type not implemented for " + serverType);
        }
        sequenceHandler.setHost(host);
        sequenceHandler.setPort(finalPort);
        sequenceHandler.setPorts(ports);
        sequenceHandler.setSocketTimeout(socketTimeout);
        sequenceHandler.setPortDelay(portDelay);
        logger.log(Level.DEBUG, "Starting server on: " + host + ":" + ports[0]);
        new Thread(new ServiceThread(sequenceHandler), "Sequence Ports Handler").start();

        //setup the server to handle the "bad" ports
        Server badPortHandler = null;
        if (serverType.equals("nio")) {
            badPortHandler = new NIOServerImpl("dummy");
        } else if (serverType.equals("aio")) {
            badPortHandler = new AsyncServerImpl("dummy");
        } else {
            throw new Exception("Server type not implemented for " + serverType);
        }
        badPortHandler.setHost(host);
        badPortHandler.setPorts(badPorts);
        badPortHandler.setSocketTimeout(socketTimeout);
        logger.log(Level.DEBUG, "Starting bad ports server on: " + host + ":" + badPorts[0]);
        new Thread(new ServiceThread(badPortHandler), "Bad Ports Handler").start();

    }

    public static void main(String[] args) throws IOException {

        Properties props = new Properties();
        try {
            props.load(new FileInputStream("daemon.properties"));

            Main main = new Main();
            main.init(props);

            //sleep until terminate signal received
            while (!terminate) {
                try {
                    Thread.currentThread().sleep(1000);
                } catch (InterruptedException ex1) {
                }
            }

        } catch (IOException ex) {
            System.err.println("Error loading properties file, server will exit.");
            //exit on error
            System.exit(1);
        } catch (Exception ex) {
            System.err.println("Error during initialization, server will exit.");
            //exit on error
            System.exit(2);
        }

        //normal exit
        System.exit(0);

    }

    /**
     * Used as a watchdog for a given server. If they stop listening prior to a
     * terminate signal, they will be placed back into the listen state.
     */
    public class ServiceThread implements Runnable {
        private Server server;

        public ServiceThread(Server server) {
            this.server = server;
        }

        public void run() {
            try {
                server.init();
                while (!terminate) {
                    server.listen();
                    logger.log(Level.DEBUG, "Listen exited for: " + server.getType());
                }
            } catch (IOException ex) {
                logger.log(Level.ERROR, "Could not start service", ex);
            }
        }

    }

}
