package org.gregoire.portknock.server;

import java.io.*;
import java.net.*;
import java.nio.channels.*;
import java.nio.channels.spi.*;
import java.util.*;
import java.util.regex.*;

import org.apache.log4j.*;

/**
 * Generic NIO server
 *
 * @author Paul Gregoire <paul%40gregoire.org>
 * @version 0.0.1
 */
public class NIOServerImpl extends Server {

    private static int counter = 0;

    private Selector selector;
    private ServerSocketChannel ssc;

    static {
        //log for this class only
        logger = Logger.getLogger(AsyncServerImpl.class.getName());
    }

    public NIOServerImpl(String type) {
        super(type);
        //list to hold all the "connectors"
        list = new LinkedList();
    }

    public void init() throws IOException {
        // Selector for incoming requests
        selector = SelectorProvider.provider().openSelector();
        // Create a new server socket and set the blocking mode
        ssc = ServerSocketChannel.open();
        ssc.configureBlocking(false);
    }

    // Accept connections for current server.
    public void listen() {
        try {
            // Bind the server socket to the local host and port or the incomming host and port
            InetSocketAddress isa = new InetSocketAddress(host, ports[0]);
            //bind the socket channel to the inet address
            ssc.socket().bind(isa);
            //set the timeout
            ssc.socket().setSoTimeout(timeout);
            // Register accepts on the server socket with the selector. This
            // step tells the selector that the socket wants to be put on the
            // ready list when accept operations occur, so allowing multiplexed
            // non-blocking I/O to take place.
            ssc.register(selector, SelectionKey.OP_ACCEPT);

            int keysAdded = 0;

            // Here's where everything happens. The select method will
            // return when any operations registered above have occurred, the
            // thread has been interrupted, etc.
            while ( (keysAdded = selector.select()) > 0) {
                // Someone is ready for I/O, get the ready keys
                Set readyKeys = selector.selectedKeys();
                Iterator i = readyKeys.iterator();
                // Walk through the ready keys collection and process requests.
                SelectionKey key = null;
                while (i.hasNext()) {
                    try {
                        key = (SelectionKey) i.next();
                        i.remove();
                        // The key indexes into the selector so you
                        // can retrieve the socket that's ready for I/O
                        SocketHandler handler = new SocketHandler();
                        Socket socket = ( (ServerSocketChannel) key.channel()).accept().socket();
                        handler.setSocket(socket);
                        new Thread(handler, "Socket handler " + counter++).start();
                        socket = null;
                    } catch (IOException e) {
                        logger.log(Level.ERROR, "Exception in listen", e);
                        key.cancel();
                    } catch (NullPointerException e) {
                        logger.log(Level.ERROR, "Exception (null) in listen", e);
                        e.printStackTrace(); // NullPointer at sun.nio.ch.WindowsSelectorImpl, Bug: 4729342
                    }

                }
                //null out iterator
                i = null;
            }

        } catch (Exception ex) {
            logger.log(Level.ERROR, "Exception in listen: " + ex.getMessage(), ex);
        }

    }

    public class SocketHandler implements Runnable {

        Socket socket = null;

        public void setSocket(Socket socket) {
            this.socket = socket;
        }

        public void run() {
            try {
                handle();
                socket.close();
            } catch (IOException ex) {
            }
        }

        /**
         * Given a Socket, this method builds the objects necessary to communicate with
         * the broadcaster and begins receiving the stream
         *
         * @param socket The socket on which the client has connected
         * @throws java.io.IOException
         */
        private void handle() throws IOException {
            InetAddress address = socket.getInetAddress();
            Client client = new Client(address);
            if (!list.contains(client)) {
                logger.log(Level.DEBUG, "Client not in list");
                list.add(client);
            } else {
                Client tmp = null;
                int listSize = list.size();
                for (int c = 0; c < listSize; c++) {
                    tmp = (Client) list.get(c);
                    if (client.equals(tmp)) {
                        //set the reference to the current client
                        client = tmp;
                        //increment the connection count
                        client.incrementSequence();
                        //if we have reached the end of the sequence then open their tunnel
                        if (client.getSequence() == ports.length) {
                            //remove the record
                            list.remove(c);
                            //open tunnel and allow only current requester in

                            logger.log(Level.DEBUG, "Got a good sequence");
                            listenFinal(client);

                            //return control
                            return;
                        }
                        break;
                    }
                    tmp = null;
                }
            }

            logger.log(Level.DEBUG, socket.getInetAddress().toString());
            logger.log(Level.DEBUG, "List size: " + list.size());

            //listen on the next port in the sequence
            listenNext(client);

        }

        // Accept connections on a follow on sequence port
        public void listenNext(Client client) {
            logger.log(Level.DEBUG, "Current port: " + ports[client.getSequence()]);
            ServerSocket serverSocket = null;
            try {
                // Bind the server socket to the local host and port or the incomming host and port
                InetSocketAddress isa = new InetSocketAddress(host, ports[client.getSequence()]);
                serverSocket = new ServerSocket();
                //bind the socket channel to the inet address
                serverSocket.bind(isa);
                //set the timeout
                serverSocket.setSoTimeout(portDelay);
                while (true) {
                    //blocks
                    socket = serverSocket.accept();
                    //check to see if this new "connector" is the one we are expecting
                    if (!client.getAddress().equals(socket.getInetAddress())) {
                        logger.log(Level.DEBUG, "Got a bad connect");
                    } else {
                        logger.log(Level.DEBUG, "Got a follow on connect");
                        handle();
                        break;
                    }
                }
            } catch (Exception ex) {
                logger.log(Level.ERROR, "Exception in listen next: " + ex.getMessage(), ex);
            } finally {
                try {
                    serverSocket.close();
                } catch (IOException ex1) {
                    logger.log(Level.ERROR, "Error closing server socket", ex1);
                }
            }

        }

        // Accept connections on a follow on sequence port
        public void listenFinal(Client client) {
            logger.log(Level.DEBUG, "Current port: " + port);
            ServerSocket serverSocket = null;
            try {
                // Bind the server socket to the local host and port or the incomming host and port
                InetSocketAddress isa = new InetSocketAddress(host, port);
                serverSocket = new ServerSocket();
                //bind the socket channel to the inet address
                serverSocket.bind(isa);
                //set the timeout
                serverSocket.setSoTimeout(portDelay);
                while (true) {
                    //blocks
                    socket = serverSocket.accept();
                    //check to see if this new "connector" is the one we are expecting
                    if (!client.getAddress().equals(socket.getInetAddress())) {
                        logger.log(Level.DEBUG, "Got a bad final connect");
                    } else {
                        logger.log(Level.DEBUG, "Got the final connect");
                        //now dump out a bunch of "ok"'s to the out stream
                        byte[] cr = "\n".getBytes();
                        byte[] out = null;

                        if (type.equals("sequence")) {
                            out = "ok ".getBytes();
                            //do stuff with socket
                            OutputStream os = socket.getOutputStream();
                            for (int w = 0; w < 1000; w++) {
                                os.write(out);
                                if (w % 10 == 0) {
                                    os.write(cr);
                                }
                            }
                            os.flush();
                        } else if (type.equals("tunnel")) {
                            String path = client.getRequestedApplication();
                            String env[] = {
                                "JAVA_HOME=E:/j2sdk1.4.2_05"};
                            Process p = Runtime.getRuntime().exec(path, env);

                            // hook processes input to browser's output (async)
                            final InputStream inFromReq = socket.getInputStream();
                            final OutputStream outToCgi = socket.getOutputStream();
                            final byte[] txBuffer = new byte[2048];

                            new Thread(new Runnable() {
                                public void run() {
                                    try {
                                        while (inFromReq.read(txBuffer) > -1) {
                                            outToCgi.write(txBuffer);
                                            outToCgi.flush();
                                        }
                                        outToCgi.close();
                                    } catch (IOException e) {
                                        logger.warn(e);
                                    }
                                }
                            }).start();
                        }
                        break;
                    }
                }
            } catch (Exception ex) {
                logger.log(Level.ERROR, "Exception in listen next: " + ex.getMessage(), ex);
            } finally {
                try {
                    serverSocket.close();
                } catch (IOException ex1) {
                    logger.log(Level.ERROR, "Error closing socket", ex1);
                }
            }

        }

    }

}
