package org.gregoire.portknock.server;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.regex.*;

import org.apache.log4j.*;

/**
 * Abstract server from which all other servers should extend.
 *
 * @author Paul Gregoire <paul%40gregoire.org>
 * @version 0.0.1
 */
public abstract class Server {

    protected String type;
    protected String host;
    protected int port;
    protected int[] ports;
    protected int timeout = 60000; //default timeout is 60s
    protected int portDelay;
    private static long clientTimeout = 120000; //2m

    private static int counter = 0;

    public static final String REGEX_IPADDRESS = "([0-9]{1,3}+" + '.' + "[0-9]{1,3}+" + '.' + "[0-9]{1,3}+" + '.' + "[0-9]{1,3}+)";

    protected static Logger logger;

    protected LinkedList list;

    public Server(String type) {
        //set the server type
        this.type = type;
        //list to hold all the "connectors"
        list = new LinkedList();
    }

    /**
     * Initializes the server.
     *
     * @throws IOException
     */
    public abstract void init() throws IOException;

    /**
     * Accept connections for current server.
     */
    public abstract void listen();

    /**
     * Sets the host name or IP address to listen on.
     *
     * @param host
     */
    public void setHost(String host) {
        this.host = host;
    }

    /**
     * Sets the final port to open. If this is a sequence server it will only open
     * the port if the sequence is completed correctly and in the specified ammount
     * of time.
     *
     * @param port
     */
    public void setPort(int port) {
        this.port = port;
    }

    /**
     * Sets the ports to open. If this is a sequence server it will only open
     * the first port initially and if it is a dummy server it will open all
     * given ports.
     *
     * @param ports
     */
    public void setPorts(int[] ports) {
        this.ports = ports;
    }

    /**
     * Sets timeout for all sockets created
     *
     * @param timeout socket timeout
     */
    public void setSocketTimeout(int timeout) {
        this.timeout = timeout;
    }

    /**
     * Sets the ammount of time in milliseconds that a client will be kept
     * active in the knock list.
     *
     * @param timeout long
     */
    public static void setClientTimeout(long cTimeout) {
        clientTimeout = cTimeout;
    }

    /**
     * Sets time that the server will listen to follow on ports.
     *
     * @param portDelay time that threads wait listening on follow on ports
     */
    public void setPortDelay(int portDelay) {
        logger.log(Level.DEBUG, "Port delay: " + portDelay);
        this.portDelay = portDelay;
    }

    /**
     * Returns the servers type, currently this consists of either "sequence" for
     * sequenced handling or "dummy" used for intrusion or port scan detection.
     *
     * @return
     */
    public String getType() {
        return type;
    }

    /**
     * Returns the ammount of time in milliseconds that a client will be kept
     * active in the knock list.
     *
     * @return long
     */
    public static long getClientTimeout() {
        return clientTimeout;
    }

    /**
     * Utility method that will return an InetAddress for a given host name or
     * IP address string.
     *
     * @param hostName host name or IP address string
     * @return InetAddress
     */
    public InetAddress getInetAddress(String hostName) {
        InetAddress address = null;
        //ip will not exceed 15 chars
        if (hostName.length() > 15) {
            try {
                address = InetAddress.getByName(hostName);
            } catch (UnknownHostException ex) {
                logger.log(Level.ERROR, "Error getting inet address by hostname", ex);
            }
        } else {
            Pattern p = Pattern.compile(REGEX_IPADDRESS);
            Matcher match = p.matcher(hostName);
            if (match.matches()) {
                int i = 0;
                byte[] ipBytes = new byte[4];
                //first check that length does not exceed xxx.xxx.xxx.xxx
                StringTokenizer st = new StringTokenizer(hostName, ".");
                while (st.hasMoreTokens()) {
                    ipBytes[i++] = (byte) Integer.parseInt(st.nextToken());
                }
                try {
                    address = InetAddress.getByAddress(ipBytes);
                } catch (UnknownHostException ex) {
                    logger.log(Level.ERROR, "Error getting inet address", ex);
                }
            } else {
                try {
                    address = InetAddress.getByName(hostName);
                } catch (UnknownHostException ex) {
                    logger.log(Level.ERROR, "Error getting inet address by hostname", ex);
                }
            }
        }
        return address;
    }

}
