package org.gregoire.portknock.util;

import java.io.*;
import java.util.*;
import java.util.regex.*;
import java.text.*;
import org.apache.log4j.*;

/**
 * Reads iptables log entries and parses the info we want for checking
 * port knocks.
 *
 * @author Paul Gregoire
 * @version 0.0.2
 */
public class FirewallLogReader {

    //match month - day - time
    private static final Pattern PAT_TIMESTAMP = Pattern.compile("([a-z|A-Z]{3}).([0-9]{1,2}).([0-9]{1,2}:[0-9]{2}:[0-9]{2}).");
    //used to match source ip address
    private static final Pattern PAT_IP_SRC = Pattern.compile("(SRC=).{7,15}[ ]{1}");
    //used to match destination ip address
    private static final Pattern PAT_IP_DEST = Pattern.compile("(DST=).{7,15}[ ]{1}");
    //used to match destination port
    private static final Pattern PAT_PORT_DEST = Pattern.compile("(DPT=).{1,5}[ ]{1}");

    //time constants
    private static final long ONE_DAY = (((1000 * 60) * 60) * 24);
    private static final long TWO_DAYS = ONE_DAY * 2;
    private static final long THREE_DAYS = ONE_DAY * 3;

    private static Logger logger;

    static {
        //log for this class only
        logger = Logger.getLogger(FirewallLogReader.class.getName());
    }

    // Sample from /var/log/messages of attempt on closed port. DPT = port attempted
    // Feb 16 15:51:05 affinity kernel: IN=eth0 OUT=
    //     MAC=00:b0:d0:b0:be:9e:00:0b:5f:c6:9c:00:08:00
    //     SRC=65.104.63.195 DST=66.209.66.70
    //     LEN=48 TOS=0x00 PREC=0x00 TTL=115 ID=13288 DF
    //     PROTO=TCP SPT=19531 DPT=82 WINDOW=64240 RES=0x00 SYN URGP=0

    public FirewallLogReader() {
    }

    public void readLog() {
        //date time format used by iptables log entries
        SimpleDateFormat formatter = new SimpleDateFormat("MMM dd kk':'mm':'ss");
        //get current data
        Calendar cal = Calendar.getInstance();
        logger.log(Level.DEBUG, "Date: " + formatter.format(cal.getTime()));
        //get current year for use later since log doesnt contain the year
        int year = cal.get(Calendar.YEAR);
        //get current time millis
        long now = cal.getTimeInMillis();
        Matcher match = null;
        BufferedReader in = null;
        try {
            in = new BufferedReader(new FileReader("messages"));
            String tmp = null, time = null, src = null, dst = null, port = null;
            while ( (tmp = in.readLine()) != null) {
                match = PAT_TIMESTAMP.matcher(tmp);
                if (match.find(0)) {
                    match.group(0);
                    time = tmp.substring(match.start(0), match.end(0));
                    //set calendar time
                    cal.setTimeInMillis(formatter.parse(time).getTime());
                    cal.set(Calendar.YEAR, year);
                    //check to see if we should grab entries
                    long logTime = cal.getTimeInMillis();
                    //System.out.println("Log: " + cal);
                    //System.out.println("Times: " + now + " " + logTime);
                    //System.out.println("Over 1 day: " + ((now - logTime) >= ONE_DAY));
                    //System.out.println("Over 2 days: " + ((now - logTime) >= TWO_DAYS));
                    //System.out.println("Over 3 days: " + ((now - logTime) >= THREE_DAYS));
                    if ((now - logTime) >= ONE_DAY) {
                        //over a day? too old
                        logger.log(Level.DEBUG, "Log entry too old, skipping");
                        continue;
                    }
                    //get source ip
                    match = PAT_IP_SRC.matcher(tmp);
                    if (match.find(0)) {
                        match.group(0);
                        src = tmp.substring(match.start(0) + 4, match.end(0));
                        //get destination ip
                        match = PAT_IP_DEST.matcher(tmp);
                        if (match.find(0)) {
                            match.group(0);
                            dst = tmp.substring(match.start(0) + 4, match.end(0));
                            //get destination port
                            match = PAT_PORT_DEST.matcher(tmp);
                            if (match.find(0)) {
                                match.group(0);
                                port = tmp.substring(match.start(0) + 4, match.end(0));

                                logger.log(Level.DEBUG, "Time: " + time + " Source: " + src + " Dest: " + dst + " Port: " + port);
                            }

                        }

                    } else {
                        logger.log(Level.DEBUG, "Non firewall entry");
                        continue;
                    }
                    //System.out.println(tmp);
                } else {
                    logger.log(Level.DEBUG, "Didnt find match");
                }
            }
        } catch (Exception ex) {
            logger.log(Level.ERROR, "Problem reading firewall log", ex);
        } finally {
            try {
                in.close();
            } catch (IOException ex) {
                logger.log(Level.ERROR, "Problem closing firewall log", ex);
            }
        }

    }

    public static void main(String[] args) {
        PropertyConfigurator.configure("bin/log4j.lcf");
        FirewallLogReader flr = new FirewallLogReader();
        flr.readLog();
    }

}